#include "HuffmanCodeLengthTable.h"

#include "ByteUtils.h"
#include "RunLengthEncoder.h"

#include <algorithm>
#include <sstream>
#include <iostream>

void HuffmanCodeLengthTable::buildCompressedLengthSequence()
{
    RunLengthEncoder rl_encoder;
    auto rle_encoded = rl_encoder.encode(mInputLengthSequence);

    for (const auto& entry : rle_encoded)
    {
        std::cout << "Got rle " << static_cast<int>(entry.first) << " | " << entry.second << std::endl;
    }
    mCompressedLengthSequence.clear();

    for (const auto& entry : rle_encoded)
    {
        const auto length = entry.first;
        const auto count = entry.second;
        if (count < 3)
        {
            for(unsigned idx=0; idx<3; idx++)
            {
                mCompressedLengthSequence.push_back({length, 0});
            }
        }
        else if (length == 0)
        {
            if(count >=3 && count <=10)
            {
                mCompressedLengthSequence.push_back({17, count-3});
            }
            else
            {
                mCompressedLengthSequence.push_back({18, count-11});
            }
        }
        else
        {
            mCompressedLengthSequence.push_back({length, 0});
            auto num_blocks_of_six = (count-1)/6;
            for(unsigned idx=0; idx<num_blocks_of_six; idx++)
            {
                mCompressedLengthSequence.push_back({16, 3});
            }
            auto remaining_counts = (count-1) % 6;
            if (remaining_counts >= 3)
            {
                mCompressedLengthSequence.push_back({16, remaining_counts - 3});
            }
            else
            {
                for(unsigned idx=0; idx<remaining_counts; idx++)
                {
                    mCompressedLengthSequence.push_back({length, 0});
                }
            }
        }
    }

    mCompressedLengthCounts = std::vector<unsigned>(19, 0);
    for (const auto& entry : mCompressedLengthSequence)
    {
        mCompressedLengthCounts[entry.first]++;
    }
}

const std::vector<HuffmanCodeLengthTable::CompressedSequenceEntry>& HuffmanCodeLengthTable::getCompressedLengthSequence() const
{
    return mCompressedLengthSequence;
}

const std::vector<unsigned> HuffmanCodeLengthTable::getCompressedLengthCounts() const
{
    return mCompressedLengthCounts;
}

void HuffmanCodeLengthTable::buildPrefixCodes()
{
    if(mInputLengthSequence.empty())
    {
        return;
    }

    unsigned char max_length = *std::max_element(mInputLengthSequence.begin(), mInputLengthSequence.end());
    std::vector<unsigned> counts(max_length+1, 0);
    for (const auto length : mInputLengthSequence)
    {
        counts[length]++;
    }
    counts[0] = 0;

    uint32_t code{0};
    std::vector<uint32_t> next_code(max_length + 1, 0);
    for (unsigned bits = 1; bits <= max_length; bits++)
    {
        code = (code + counts[bits-1]) << 1;
        next_code[bits] = code;
    }

    for(std::size_t idx=0; idx<mInputLengthSequence.size(); idx++)
    {
        if (const auto length = mInputLengthSequence[idx]; length != 0)
        {
            const auto code = next_code[length];
            next_code[length]++;
            auto prefix_code = PrefixCode(code, length);
            mTree.addCodeLengthEntry(length, {PrefixCode(code, length), idx});
            mCodes.push_back(prefix_code);
        }
    }
    mTree.sortTable();
    //std::cout << dumpPrefixCodes();
}

const PrefixCode& HuffmanCodeLengthTable::getCode(std::size_t index) const
{
    return mCodes[index];
}

std::string HuffmanCodeLengthTable::dumpPrefixCodes() const
{
    return mTree.dump();
}

unsigned HuffmanCodeLengthTable::mapToDeflateIndex(unsigned index) const
{
    if (index>= DEFLATE_PERMUTATION_SIZE)
    {
        return 0;
    }
    else
    {
        return DEFLATE_PERMUTATION[index];
    }
}

unsigned HuffmanCodeLengthTable::getNumCodeLengths() const
{
    return mTree.getNumCodeLengths();
}

std::optional<HuffmanTree::Symbol> HuffmanCodeLengthTable::findMatch(std::size_t treeIndex, uint32_t code) const
{
    return mTree.findMatch(treeIndex, code);
}

unsigned HuffmanCodeLengthTable::getCodeLength(std::size_t index) const
{
    return mTree.getCodeLength(index);
}

void HuffmanCodeLengthTable::setInputLengthSequence(const std::vector<unsigned char>& sequence, bool targetDeflate)
{
    mTargetDeflate = targetDeflate;

    if (targetDeflate)
    {
        mInputLengthSequence = std::vector<unsigned char>(DEFLATE_PERMUTATION_SIZE, 0);
        for(unsigned idx=0; idx<sequence.size(); idx++)
        {
            mInputLengthSequence[mapToDeflateIndex(idx)] = sequence[idx];
            //std::cout << "Got code length for " << mapToDeflateIndex(idx) << " of " << static_cast<unsigned>(sequence[idx]) << std::endl;
        }
    }
    else
    {
        mInputLengthSequence = sequence;
    }
}
